import json
import os
import pickle
from itertools import islice

from oireachtas_data.constants import DEBATES_DIR
from oireachtas_data.models.debate import Debate
from oireachtas_data.models.para import Para


def find_nth(haystack, needle, n):
    start = haystack.find(needle)
    while start >= 0 and n > 1:
        start = haystack.find(needle, start+len(needle))
        n -= 1
    return start


def merge_paras(paras):
    '''
    When you don't care about context it's often easier to treat many
    paragraphs as one.

    :param paras:
    :return: A Para object that just merges all the paras you give it into one.
    '''
    return Para(content='\n\n'.join([m.content for m in paras]))


def get_file_content(filepath):
    """
    Read a file regardless of the extension

    For instance if there's some pickle files and some pickle.bz files
    this will give back the same data as if they were not compressed

    :param filepath: Filepath to read from
    :return: Depending on the ext it may be a pickled obj / str
    """
    ext = os.path.splitext(filepath)[1]
    if ext == '.txt':
        return open(filepath, 'r').read()
    elif ext == '.json':
        f = open(filepath, 'r')
        data = json.loads(f.read())
        f.close()
        return data
    elif ext == '.pickle':
        return pickle.load(open(filepath, 'rb'))
    else:
        raise NotImplementedError()


def iter_debates():
    for f in os.listdir(DEBATES_DIR):
        if f.endswith('.json'):
            yield Debate.from_file(
                os.path.join(
                    DEBATES_DIR,
                    f
                )
            )


def get_debates():
    debates = []
    for f in os.listdir(DEBATES_DIR):
        if f.endswith('.json'):
            debates.append(
                Debate.from_file(
                    os.path.join(
                        DEBATES_DIR,
                        f
                    )
                )
            )
    return debates


def window(sequence, window_size=2):
    """
    Returns a sliding window (of width n) over data from the iterable
    """
    seq_iterator = iter(sequence)
    result = tuple(islice(seq_iterator, window_size))
    if len(result) == window_size:
        yield result
    for elem in seq_iterator:
        result = result[1:] + (elem,)
        yield result


def first_occuring(strings, content):
    earliest = None
    earliest_idx = float('inf')
    for s in strings:
        try:
            idx = content.index(s)
            if idx < earliest_idx:
                earliest = s
                earliest_idx = idx
        except ValueError:
            pass

    return (earliest_idx, earliest)



import re
import os
import datetime
import string
from difflib import SequenceMatcher

from cached_property import cached_property

from oireachtas_data.models.speech import Speech
from oireachtas_data.models.para import Para


class Section():

    def __init__(self, data=None, title=None, next_title=None):
        self.data = data
        self.title = title
        self.next_title = next_title

    @property
    def content(self):
        from oireachtas_data.utils import first_occuring

        lines = self.non_header_content.split('\n')

        valid_lines = []
        for line in lines:
            line = line.strip()
            good = True

            try:
                int(line[0])
                datetime.datetime.strptime(
                    line.strip()[:10],
                    '%d/%m/%Y'
                )
                good = False
            except (ValueError, IndexError):
                pass

            if good:
                valid_lines.append(line)

        content = '\n'.join(valid_lines)

        data = ''

        if self.next_title is None:
            if self.title in content:
                data = content[
                    content.index(self.title) + len(self.title) + 1:
                ]
            elif self.title.rstrip(string.digits) in content:
                data = content[
                    content.index(self.title.rstrip(string.digits)) + len(self.title.rstrip(string.digits)) + 1:
                ]
            elif len(self.title) > 50 and self.title[:40] in content:
                data = content[
                    content.index(self.title[:40]) + len(self.title[:40]):
                ]

            else:
                # This could be that the title down below is so long that it's split
                print('oops, could not find the last section')

        else:
            if self.title not in content or self.next_title not in content:

                lines = content.split('\n')

                start_index = None
                end_index = None

                if self.title not in content:

                    modified_title = self.title.rstrip(string.digits)
                    if modified_title in content:
                        start_index = content.index(modified_title)
                    else:
                        from oireachtas_data.utils import window
                        for p, n in window(lines, window_size=2):
                            if p + ' ' + n == modified_title:
                                start_index = content.index(p + '\n' + n)
                else:
                    start_index = content.index(self.title)

                if start_index is None:
                    # try remove numbers from the end of the title
                    print(f'Could not get {self.title} from content')
                    return ''

                if self.next_title not in content:
                    from oireachtas_data.utils import window
                    for p, n in window(lines, window_size=2):
                        if p + ' ' + n == self.next_title:
                            end_index = content.index(p + '\n' + n)
                else:
                    end_index = content.index(self.next_title)

                data = content[
                    start_index + len(self.title) + 1:end_index
                ]
            else:
                data = content[
                    content.index(self.title) + len(self.title) + 1:content.index(self.next_title)
                ]

        # Remove tables of names from voting
        while 'The Dáil divided: Tá,' in data or 'The Committee divided: Tá,' in data or 'The Seanad divided: Tá,' in data:
            earliest = first_occuring(['The Dáil divided: Tá,', 'The Committee divided: Tá,', 'The Seanad divided: Tá,'], data)
            first_occuring_string = earliest[1]

            table_start = data.index(first_occuring_string)

            earliest = first_occuring(['Question declared carried.', 'Amendment declared lost.'], data[table_start:])
            if earliest[1] is None:
                break

            table_end = earliest[0]

            start = data[:table_start]
            end = data[table_start + table_end + 5:]

            data = start + '\n\n' + end

        data = ' '.join(data.split('[PAGEBREAK]'))

        return data.strip()

    def __del__(self):
        attrs_to_del = [
            'non_header_content'
        ]

        for attr in attrs_to_del:
            try:
                delattr(self, attr)
            except AttributeError:
                pass

    @cached_property
    def non_header_content(self):
        return self.data[
            self.data.rindex('����������������������������������'):
        ]

    @property
    def speeches(self):
        if ':' not in self.content:
            return []

        data = self.content

        lines = data.split('\n')
        new_lines = []
        for line in lines:
            line = line.strip()

            if ':' in line and any([
                'Deputy' in line,
                line[:line.index(':')].strip().split(' ')[-1][0].isupper()
            ]):
                new_lines.append(line)
            else:
                if len(new_lines):
                    new_lines[-1] = new_lines[-1] + ' ' + line

        speeches = []
        for line in new_lines:
            person = line[0:line.index(':')].strip()
            content = line[line.index(':') + 1:]
            content = content.replace('   Seanad Éireann  ', ' ').strip()

            speeches.append(
                Speech(
                    by=person,
                    _as=None,
                    eid=None,
                    paras=[Para(title=None, eid=None, content=content)]
                )
            )

        return speeches


class PDF():

    def __init__(self, fp):
        self.fp = fp
        self.data = None
        self.loaded = False

    def load(self):
        text_file = self.fp + '.txt'

        if not os.path.exists(text_file):
            with open(self.fp, 'rb') as f:
                os.system(f'pdftotext \'{self.fp}\' \'{text_file}\'')

        if not os.path.exists(text_file):
            print('Could not convert pdf to text: %s' % (text_file))
            return

        f = open(text_file, 'r')
        self.data = f.read()
        f.close()

        # Just weird thigs to remove, probably wobblyness in conversion to text
        self.data = self.data.replace('-----', ' — ')  # interruption / continuation of interruption
        self.data = self.data.replace('\x0c', '')

        # Removes timestamps of speach
        self.data = re.sub('\\d+\\/\\d+\\/\\d+[A-Za-z]+\\d+', '', self.data)

        self.data = self.data.replace('An Leas-Chathaoirleach:', '\nAn Leas-Chathaoirleach:')
        self.data = self.data.replace('----An Leas-Chathaoirleach:', '\nAn Leas-Chathaoirleach:')
        self.data = self.data.replace('----An Cathaoirleach:', '\nAn Cathaoirleach:')
        self.data = self.data.replace('----An Ceann Comhairle:', '\nAn Ceann Comhairle:')
        self.data = self.data.replace('----Deputy', '\nDeputy')
        self.data = self.data.replace('----Senator', '\nSenator')
        self.data = self.data.replace('----Acting Chairman', '\nActing Chairman')

        # would be handy to replace all "An Leas-Chathaoirleach:" with "\nAn Leas-Chathaoirleach:" so we can parse things easier really, sometimes they get merged into other words (and other common names)

        lines = self.data.split('\n')
        good_lines = []

        for line in lines:
            line = line.strip()
            good = True
            try:
                int(line[0])
                datetime.datetime.strptime(
                    line.strip(),
                    '%d %B %Y'
                )
                good = False
            except:
                pass

            if line.strip() == 'Dáil Éireann':
                good = False

            if good:
                good_lines.append(line)

        lines = good_lines
        good_lines = []

        for line in lines:
            line = line.strip()
            try:
                int(line)
                good_lines.append('[PAGEBREAK]')
            except (ValueError, IndexError):
                good_lines.append(line)

        self.data = '\n'.join(good_lines)

    def matching_header(self, show_as):
        for header in self.section_headers:
            if SequenceMatcher(None, header, show_as).ratio() > 0.8:
                return header
        return None

    @property
    def section_headers(self):
        if not self.loaded:
            self.load()

        if self.data is None:
            return []

        lines = self.data.split('\n')
        start = -1
        end = -1
        for idx, line in enumerate(lines):
            # If there's a continue then it may be skipped. Long lines
            # get first and last of ����� then if no ����� in line merge with the next line
            if '��������������' in line and start != -1:
                end = idx
            if '��������������' in line and start == -1:
                start = idx

        header_lines = lines[start:end+1]
        cleaned_header_lines = []
        append_next = False
        for header_line in header_lines:
            if '\x08' in header_line:
                if append_next:
                    cleaned_header_lines[-1] += ' ' + header_line[:header_line.index('�')]
                else:
                    if '�' not in header_line:
                        # page number at end
                        cleaned_header_lines.append(header_line)
                    else:
                        cleaned_header_lines.append(header_line[:header_line.index('�')])
                append_next = False
            else:
                cleaned_header_lines.append(header_line)
                append_next = True

        clean_headers = []
        for header in cleaned_header_lines:

            header = header.replace('\x08', '')
            if len(header.split()):
                if header.split()[-1].isnumeric():
                    header = ' '.join(header.split()[:-1])

            header = header.strip()
            if '\x08' in header:
                header = header[:header.index('\x08')].replace('\n', ' ')

            header = ' '.join(header.strip().split('\n')).strip()

            if header:
                try:
                    int(header[0])
                    datetime.datetime.strptime(header, '%d/%m/%YA%M%S0')
                except (ValueError, IndexError):
                    clean_headers.append(header)
        return clean_headers

    @property
    def date(self):
        return datetime.datetime.strptime(
            os.path.splitext(self.fp.split('_')[-1])[0],
            '%Y-%m-%d'
        )

    @property
    def debate_sections(self):
        if not self.loaded:
            self.load()
        from oireachtas_data.utils import window
        sections = []
        for title, next_title in window(self.section_headers + [None], window_size=2):
            sections.append(
                Section(
                    data=self.data,
                    title=title,
                    next_title=next_title
                )
            )
        return sections



from unittest import TestCase, skip

from oireachtas_data.utils import (
    merge_paras,
    get_file_content,
    get_debates,
    window,
    first_occuring
)


class UtilsTest(TestCase):

    @skip('TODO')
    def test_merge_paras(self):
        pass

    @skip('TODO')
    def test_get_file_content(self):
        pass

    @skip('TODO')
    def test_get_debates(self):
        pass

    def test_window(self):
        self.assertEqual(
            list(window(range(10), window_size=2)),
            [(0, 1), (1, 2), (2, 3), (3, 4), (4, 5), (5, 6), (6, 7), (7, 8), (8, 9)]
        )

    def test_first_occuring(self):
        self.assertEqual(
            first_occuring(
                ['three', 'two'],
                'one two three four'
            ),
            (4, 'two')
        )



import argparse
import datetime
import json
import os
from multiprocessing.pool import ThreadPool

import tqdm

from urllib.request import Request
from urllib.request import urlopen

from oireachtas_data.models.debate import Debate
from oireachtas_data.constants import DEBATES_DIR
from oireachtas_data.utils import get_debates


def scrape_debates(d):
    try:
        d.load_data()
        d.write()
    except:
        print('Possibly wrong pdf parsed: %s - %s' % (d.date, d.chamber))


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--processes',
        type=int,
        dest='num_processes',
        default=20
    )
    args = parser.parse_args()

    if not os.path.exists(DEBATES_DIR):
        os.makedirs(DEBATES_DIR, exist_ok=True)

    earliest_possible = datetime.datetime(1900, 1, 1)
    chamber_types = ['house', 'committee']
    try:
        earliest = min([d.timestamp for d in get_debates()])
    except:
        earliest = datetime.datetime.now()

    # There are a lot of access denieds, should look here for missing data
    # https://www.oireachtas.ie/en/debates/debate/seanad/{x}/{y}/ from https://data.oireachtas.ie/akn/ie/debateRecord/seanad/{x}/debate/mul@/dbsect_{y}.xml
    # will need to parse from the page which will be anoying

    for chamber_type in chamber_types:
        while earliest > earliest_possible:
            earliest = earliest.replace(day=1)

            chamber_id = ''
            chamber = ''
            date_start = earliest.strftime('%Y-%m-%d')
            date_end = (
                (earliest + datetime.timedelta(days=32)).replace(day=1) - datetime.timedelta(days=1)
            ).strftime('%Y-%m-%d')
            limit = 1000

            print(
                'Get debates of config: %s' % (
                    {
                        'chamber_id': chamber_id,
                        'chamber': chamber,
                        'chamber_type': chamber_type,
                        'date_start': date_start,
                        'date_end': date_end,
                        'limit': limit,
                    }
                )
            )

            req = Request(
                'https://api.oireachtas.ie/v1/debates?chamber_type=%s&chamber_id=%s&chamber=%s&date_start=%s&date_end=%s&limit=%s' % (
                    chamber_type,
                    chamber_id,
                    chamber,
                    date_start,
                    date_end,
                    limit
                ),
                headers={'accept': 'application/json'}
            )
            response = urlopen(req)
            data = json.loads(response.read())

            print('got debates: %s' % (len(data['results'])))

            debates = []
            for d in data['results']:
                debates.append(
                    Debate(
                        date=d['contextDate'],
                        chamber=d['debateRecord']['chamber']['showAs'],
                        counts=d['debateRecord']['counts'],
                        debate_type=d['debateRecord']['debateType'],
                        debate_sections=d['debateRecord']['debateSections']
                    )
                )

            pool = ThreadPool(processes=args.num_processes)
            for _ in tqdm.tqdm(
                pool.imap_unordered(scrape_debates, debates),
                total=len(debates)
            ):
                pass

            earliest = earliest - datetime.timedelta(days=1)


if __name__ == '__main__':
    main()



import urllib
from urllib.request import urlopen

import bs4

from oireachtas_data.models.speech import Speech
from oireachtas_data.models.para import Para
from oireachtas_data.models.pdf import pdf_parser_get


pdf_cache = {}


class DebateSection():

    __slots__ = (
        'bill',
        'contains_debate',
        'counts',
        'debate_section_id',
        'debate_type',
        'data_uri',
        'parent_debate_section',
        'show_as',
        'show_as_idx',
        'show_as_context',
        'speakers',
        'speeches',
        'pdf_location',
        'loaded'
    )

    def __init__(
        self,
        bill=None,
        contains_debate=None,
        counts=None,
        debate_section_id=None,
        debate_type=None,
        data_uri=None,
        parent_debate_section=None,
        show_as=None,
        show_as_idx=None,
        show_as_context=None,
        speakers=None,
        speeches=None,
        pdf_location=None,
        loaded=False
    ):
        self.bill = bill
        self.contains_debate = contains_debate
        self.counts = counts
        self.debate_section_id = debate_section_id
        self.debate_type = debate_type
        self.data_uri = data_uri
        self.parent_debate_section = parent_debate_section
        self.show_as = show_as.strip()
        self.show_as_idx = show_as_idx
        self.show_as_context = show_as_context
        self.speakers = speakers
        self.speeches = speeches if speeches else []

        self.pdf_location = pdf_location
        self.loaded = loaded

    @staticmethod
    def parse(data):
        return DebateSection(
            bill=data['bill'],
            contains_debate=data['contains_debate'],
            counts=data['counts'],
            debate_section_id=data['debate_section_id'],
            debate_type=data['debate_type'],
            data_uri=data['data_uri'],
            parent_debate_section=data['parent_debate_section'],
            show_as=data['show_as'],
            speakers=data['speakers'],
            speeches=[
                Speech.parse(s) for s in data['speeches']
            ],
            loaded=True
        )

    def load_data(self):
        '''
        Load data from the data_uri to populate the speeches and
        their paragraphs
        '''
        if self.loaded:
            return

        try:
            source = urlopen(self.data_uri)
        except (Exception, urllib.error.HTTPError) as ex:
            if str(ex) != 'HTTP Error 403: Forbidden':
                raise ex
            else:
                parser = pdf_parser_get(self.pdf_location)

                if parser is None:
                    return

                if self.pdf_location not in pdf_cache:
                    pdf_cache[self.pdf_location] = parser(self.pdf_location)

                pdf = pdf_cache[self.pdf_location]

                # Need to know what occurace of the show_as is, since there can be multiples
                # will then need to get that one in the debate section

                # this self.show_as is the xth occurance
                #xth_occurance = self.show_as_context[0:self.show_as_idx + 1].count(self.show_as)
                #import pdb; pdb.set_trace()

                matching_header = pdf.matching_header(self.show_as)
                if matching_header is None:
                    # FIXME: may not be in debate sections, could just be on its own
                    # look if there's a line just containing the header
                    print(f'Could not find {self.show_as} in {self.pdf_location}')
                else:
                    # TODO: as sections are used, remove them as there can be multples, don't want to repeat
                    # can update the model since it's shared

                    try:
                        section = [s for s in pdf.debate_sections if s.title.lower().strip() == matching_header.lower().strip()][0]
                        self.speeches.extend(
                            section.speeches
                        )
                    except IndexError:
                        print('Could not get header "%s" from "%s"' % (matching_header, pdf.fp))

                return

        soup = bs4.BeautifulSoup(source, 'html.parser')

        # heading
        # soup.find('debatesection').find('heading').text

        for speech in soup.find('debatesection').find_all('speech'):
            paras = [
                Para(
                    title=p.attrs.get('title', None),
                    eid=p.attrs.get('eid', None),
                    content=p.text
                ) for p in speech.find_all('p')
            ]

            self.speeches.append(
                Speech(
                    by=speech.attrs.get('by'),
                    _as=speech.attrs.get('as'),
                    eid=speech.attrs.get('eid'),
                    paras=paras
                )
            )

        # Could also get summary

        self.loaded = True

    def serialize(self):
        return {
            'bill': self.bill,
            'contains_debate': self.contains_debate,
            'counts': self.counts,
            'debate_section_id': self.debate_section_id,
            'debate_type': self.debate_type,
            'data_uri': self.data_uri,
            'parent_debate_section': self.parent_debate_section,
            'show_as': self.show_as,
            'speakers': self.speakers,
            'speeches': [s.serialize() for s in self.speeches]
        }

    def speech_contains(self, substr, case_sensitive=True):
        content = ''
        for speech in self.speeches:
            content += '\n' + speech.content

        if case_sensitive:
            return substr in content
        else:
            return substr.lower() in content.lower()

    @property
    def content(self):
        return '\n'.join([
            '\n' + speech.by + ': ' + speech.content for speech in self.speeches
        ])

    @property
    def is_from_pdf(self):
        return self.debate_type is None

    @property
    def is_empty(self):
        # this can happen when there's access denied errors
        return self.speakers == [] and self.speeches == []

    @property
    def computed_speakers(self):
        speakers = set()
        for speech in self.speeches:
            speakers.add(speech.by)

        return speakers



import os

from oireachtas_data.models.new_pdf import PDF as NEW_PDF
from oireachtas_data.models.old_pdf import PDF as OLD_PDF


def to_text(fp):
    text_file = fp + '.txt'

    if not os.path.exists(text_file):
        with open(fp, 'rb') as f:
            os.system(f'pdftotext \'{fp}\' \'{text_file}\'')

    if not os.path.exists(text_file):
        print('Could not convert pdf to text: %s' % (text_file))
        return

    f = open(text_file, 'r')
    data = f.read()
    f.close()

    return data


def pdf_parser_get(pdf_location):

    content = to_text(pdf_location)

    if content is None:
        print('Could not parse pdf type: %s' % (pdf_location))
        return

    if '��������������' in content:
        return NEW_PDF
    return OLD_PDF



from setuptools import (
    find_packages,
    setup
)

INSTALL_REQUIRES = (
    'requests',
    'cached_property',
    'bs4',
    'nltk',
    'pdftotext',
    'transformers',
    'unidecode',
    'edlib'
)

setup(
    name='oireachtas-data',
    version='0.1.3',
    python_requires='>=3.6',
    description='Oireachtas debate data',
    long_description='Oireachtas debate data',
    author='Robert Lucey',
    url='https://github.com/RobertLucey/oireachtas-data',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    install_requires=INSTALL_REQUIRES,
    entry_points={
        'console_scripts': [
            'load_debates = oireachtas_data.bin.load_debates:main',
            'pull_debates = oireachtas_data.bin.pull_debates:main',
        ]
    }
)



from oireachtas_data.models.para import Para

from oireachtas_data import members


class Speech():

    __slots__ = (
        'by',
        '_as',
        'eid',
        'paras'
    )

    def __init__(self, by=None, _as=None, eid=None, paras=None):
        '''

        :kwarg by: The name / identifier of the speaker
        :kwarg _as: The title / position of the speaker
        :kwarf eid: incrementing id of the speech
        :kwarg paras: list of Para objects
        '''
        self.by = by
        self._as = _as
        self.eid = eid

        if paras is not None and paras != []:
            types = list(set([type(p) for p in paras]))
            assert(len(types) == 1)
            assert(types[0] == Para)

        self.paras = paras if paras else []

    @staticmethod
    def parse(data):
        return Speech(
            by=data['by'],
            _as=data['as'],
            eid=data['eid'],
            paras=[Para.parse(p) for p in data['paras']]
        )

    def serialize(self):
        return {
            'by': self.by,
            'as': self._as,
            'eid': self.eid,
            'paras': [p.serialize() for p in self.paras]
        }

    @property
    def content(self):
        return '\n'.join([
            '\n' + para.content for para in self.paras
        ])

    @property
    def member_obj(self):
        return members.get_member(self.by)



import nltk

from oireachtas_data.models.member import Members

members = Members()
members.load()


try:
    nltk.data.find('taggers/averaged_perceptron_tagger')
except LookupError:
    nltk.download('averaged_perceptron_tagger')

try:
    nltk.data.find('tokenizers/punkt')
except LookupError:
    nltk.download('punkt')



import string
from collections import Counter

from cached_property import cached_property

import nltk
from nltk.tag.perceptron import PerceptronTagger

TAGGER = PerceptronTagger()


class Paras():

    def __init__(self, data=None):
        self.data = data

class Para():
    '''
    Paragraph object, has a title, id, and content.

    Inherits from Text which has all the text analysis bits
    '''

    __slots__ = (
        'title',
        'eid',
        'content'
    )

    def __init__(self, *args, **kwargs):
        '''

        :kwarg title: Sometimes a paragraph has a title, not often
        :kwarg eid: Incremented id of the paragraph
        :kwarg content: The str text content
        '''
        self.title = kwargs.get('title', None)
        self.eid = kwargs.get('eid', None)
        self.content = kwargs.get('content', None)

    @property
    def __dict__(self):
        return self.serialize()

    @staticmethod
    def parse(data):
        return Para(
            title=data['title'],
            eid=data['eid'],
            content=data['content']
        )

    @cached_property
    def words(self):
        '''

        :return: A list of nltk tokenized words
        '''
        return nltk.word_tokenize(self.content)

    @cached_property
    def tokens(self):
        '''

        :reuturn: A list of nltk tokenized words with their tag
        '''
        return TAGGER.tag(self.words)

    def serialize(self):
        return {
            'title': self.title,
            'eid': self.eid,
            'content': self.content
        }
